----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    14:21:05 10/19/2012 
-- Design Name: 
-- Module Name:    tscj3 - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

library WORK;
use WORK.common.ALL;



entity tscj3 is
	Port (
		input 		: in t_tscj3_i;
		output 		: out t_tscj3_o;
		clk40 		: in std_logic
		);
end tscj3;

architecture Behavioral of tscj3 is
	signal i	: t_tscj3_i;
	signal o : t_tscj3_o := t_tscj3_o_init;
	
	signal fresh : unsigned(fresh_bits-1 downto 0) := (others => '1');
	
	signal ECR			: std_logic_vector(2 downto 0)	:= (others => '0');
	signal ECRsim		: std_logic_vector(2 downto 0)	:= (others => '0');

	signal fifo_reset	: std_logic 							:= '1'; -- para vaciar la fifo cuando haya un gsr
	signal fifo_wr		: std_logic 							:= '0';
	signal fifo_rd		: std_logic 							:= '0';
	signal fifo_din	: std_logic_vector(19 downto 0)	:= (others => '0');
	signal fifo_dout	: std_logic_vector(19 downto 0);
	signal fifo_full	: std_logic;
	signal fifo_empty	: std_logic;
	
	signal refcount_data : std_logic_vector(12 downto 0) := (others => '0');

	-- Serial word strings:
	constant start   		: std_logic 			:= '0';
	constant stop   		: std_logic 			:= '1';
	constant idleword   	: std_logic_vector 	:= "00000000001111111111";
	constant resetted  	: std_logic_vector 	:= "0" & "00110" & "0000000000000" & "1";
	constant header		: std_logic_vector 	:= "0001";
	constant data   		: std_logic_vector 	:= "01";
	constant data2  		: std_logic_vector 	:= "10";
	constant trailer  	: std_logic_vector 	:= "0010";
	constant refcount 	: std_logic_vector 	:= "00111";

	COMPONENT tsc_fifo
	  PORT (
		 rst : IN STD_LOGIC;
		 wr_clk : IN STD_LOGIC;
		 rd_clk : IN STD_LOGIC;
		 din : IN STD_LOGIC_VECTOR(19 DOWNTO 0);
		 wr_en : IN STD_LOGIC;
		 rd_en : IN STD_LOGIC;
		 dout : OUT STD_LOGIC_VECTOR(19 DOWNTO 0);
		 full : OUT STD_LOGIC;
		 empty : OUT STD_LOGIC
	  );
	END COMPONENT;

	
begin
i <= input;
output <= o;

	o.debug(1)  <= '0';
	o.debug(2)  <= ECR(0);
	o.debug(3)  <= '0';
	o.debug(4)  <= ECR(1);
	o.debug(5)  <= '0';
	o.debug(6)  <= ECR(2);
	o.debug(7)  <= '0';
	o.debug(8)  <= i.evcntres;
	o.debug(9)  <= '0';
	o.debug(10) <= i.reg(ow1_ctrl_reg)(5) and ECR(1) and not ECR(0);
	o.debug(11) <= '0';
	o.debug(12) <= '0';
	o.debug(13) <= '0';
	o.debug(14) <= '0';
	o.debug(15) <= '0';
	o.debug(16) <= '0';
	o.debug(17) <= '0';
	o.debug(18) <= i.trg_strobe;
--	o.debug(19) <= reset;
--	o.debug(20) <= clk40;



process (clk40)
	variable ec_next_l1a 	: unsigned(11 downto 0) 			:= (others => '0'); -- old eventcnt; indicates the ec that will correspond to the next l1a to be received
	variable ec_next_tsc		: unsigned(11 downto 0) 			:= (others => '0'); -- old next_ec; indicates the ec that will correspond to the next ros_ready
	
	variable sincelastl1a : integer range 255 downto 0 := 0;
	variable l1a_pipe : std_logic_vector(3 downto 0) 	:= (others => '0');
	
	variable wordcount : unsigned(7 downto 0) := (others => '0');
	variable timeoutcount : integer range 255 downto 0;
	
	
	type status_t is (idle, ready, receiving);
	variable status : status_t := idle;
	
	constant idlewait : integer := 5;
	variable idling : integer range idlewait downto 0 := 0;

	variable error_register 		: std_logic_vector(23 downto 0) := (others => '0');
	variable error_register_la 	: std_logic_vector(23 downto 0) := (others => '0');
	variable debug	: std_logic_vector(3 downto 0);


begin
	if rising_edge(clk40) then
		if fresh > 0 then 
			fresh <= fresh - 1;
		end if;
		
		error_register(err_ixo_lcl) 	:= i.ixo_on;
		error_register(err_bufpll) 	:= not i.bufpll_locked;
		error_register(err_pll) 		:= not i.pll_locked;
		error_register(err_fifo) 		:= fifo_full;
		
		if i.evcntres = '1' or i.ext_wr(tsc_status_0_la) = '1' then -- Si event counter reset, resetear y pasar a idle.
			ec_next_l1a := (others => '0');
			ec_next_tsc := (others => '0');
			status := idle;
			idling := idlewait;
			l1a_pipe := (others => '0');
		elsif l1a_pipe(3 downto 2) = "01" then 
			ec_next_l1a := ec_next_l1a + 1;
			sincelastl1a := 0;
		elsif sincelastl1a < 255 then
			sincelastl1a := sincelastl1a + 1;
		end if;
		
		if sincelastl1a > to_integer(unsigned(i.reg(tsc_resyncgap))) then refcount_data <= ("1" & std_logic_vector(ec_next_l1a));
		else refcount_data <= ("0" & std_logic_vector(ec_next_l1a));
		end if;
		
		l1a_pipe := l1a_pipe(l1a_pipe'high-1 downto 0) & i.l1a;
		
		ECR <= i.evcntres & (ECR(2) or (i.reg(ow1_ctrl_reg)(5) and ECR(1) and not ECR(0))) & ECR(1);
		ECRsim <= i.evcntres & (ECRsim(2) or (i.reg(ow1_ctrl_reg)(5) and ECRsim(1) and not ECRsim(0))) & ECRsim(1);
		fifo_reset <= i.evcntres or i.ext_wr(tsc_status_0_la);
		fifo_wr <= '0';
		
		case status is
			when idle => -- Respeta el tiempo mnimo y espera a que haya triggers en cola
				o.ros_ready <= '0';
				if idling > 0 then idling := idling - 1;
				elsif ec_next_tsc /= ec_next_l1a then
					o.ros_ready <= '1';
					fifo_din <= start & header & "00" & std_logic_vector(ec_next_tsc) & stop;
					fifo_wr <= '1';
					ec_next_tsc := ec_next_tsc + 1;
					timeoutcount := to_integer(unsigned(i.reg(tsc_strobewait)));
					wordcount := (others => '0');
					status := ready;
					idling := idlewait;
				end if;
			when ready => --espera a que se levante trg_strobe y manda la primera palabra
				if timeoutcount = 0 then error_register(err_trg_strobe) 	:= '1';
				else timeoutcount := timeoutcount - 1;
				end if;
				if i.trg_strobe = '1'  then
					fifo_din <= start & data & i.trg_data & stop;
					fifo_wr <= '1';
					status := receiving;
				end if;
			when receiving => --manda palabras hasta que termina trg_strobe o hay un wordcountmax
				wordcount := wordcount + 1;
				if i.trg_strobe = '1' and wordcount < to_integer(unsigned(i.reg(tsc_maxwords))) then
					fifo_din <= start & data & i.trg_data & stop;
					fifo_wr <= '1';
				else
					fifo_din <= start & trailer & "000000" & std_logic_vector(wordcount) & stop;
					fifo_wr <= '1';
					status := idle;
					if wordcount = to_integer(unsigned(i.reg(tsc_maxwords))) then error_register(err_dat_count_max) := '1';
					end if;
				end if;
		end case;

		-- actualizar registro de flags de error y latched.
		if i.ext_wr(tsc_status_1_la) = '1' or fresh = 1 then error_register_la := error_register;
		else error_register_la := error_register_la or error_register;
		end if;
		error_register := (others => '0');
		
		o.reg_int(tsc_status_0_la) 	<= error_register_la(7 downto 0);
		o.reg_int(tsc_status_1_la) 	<= error_register_la(15 downto 8);
		o.reg_int(tsc_status_2_la) 	<= error_register_la(23 downto 16);
		o.reg_ch_int(tsc_status_0_la) 	<= not i.reg_ch_int(tsc_status_0_la);
		o.reg_ch_int(tsc_status_1_la) 	<= not i.reg_ch_int(tsc_status_1_la);
		o.reg_ch_int(tsc_status_2_la) 	<= not i.reg_ch_int(tsc_status_2_la);

	end if;
end process;

tsc_fifo_inst : tsc_fifo
	PORT MAP (
		wr_clk => clk40,
		rd_clk => i.clk40s,
		rst => fifo_reset,
		din => fifo_din,
		wr_en => fifo_wr,
		rd_en => fifo_rd,
		dout => fifo_dout,
		full => fifo_full,
		empty => fifo_empty
		);

process (i.clk40s)
	variable justclosed : std_logic := '0';
	variable counter		: unsigned(9 downto 0) := (others => '1');
begin
	if rising_edge(i.clk40s) then
		fifo_rd <= '0';
		if ECR(0) = '1' then
			o.word <= resetted;
			justclosed := '0';
			counter := (others => '1');
		elsif ECRsim(0) = '1' then 
			o.word <= idleword;
			justclosed := '0';
			counter := (others => '1');
		elsif (justclosed = '1' or counter = 0) and refcount_data(12) = '1' then
			o.word <= start & refcount & refcount_data & stop;
			justclosed := '0';
			counter := (others => '1');
		elsif fifo_empty = '0' then
			fifo_rd <= '1';
			counter := (others => '1');
			if fifo_rd = '1' then 
				o.word <= fifo_dout;
				if fifo_dout(18 downto 17) = data and i.trg_strobe = '0' then o.word(18 downto 17) <= data2;
				end if;
				if fifo_dout(18 downto 15) = trailer then justclosed := '1';
				end if;
			else o.word <= idleword;
			end if;
		else
			o.word <= idleword;
			if counter > 0 then counter := counter - 1;
			end if;
		end if;
	end if;
end process;

end Behavioral;

