----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    14:40:52 07/09/2012 
-- Design Name: 
-- Module Name:    onewire - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if itantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

library WORK;
use WORK.common.ALL;

entity onewire is
	Generic (
		ctrl_reg		: integer;
		read_reg		: integer;
		write_reg	: integer;
	   crc_reg		: integer
	);
   Port (
		input  		: in t_onewire_i;
		output 		: out t_onewire_o;
		wire			: inout std_logic := 'Z';
		clk40 		: in std_logic
	 );
end onewire;

architecture Behavioral of onewire is

	signal i	: t_onewire_i;
	signal o : t_onewire_o := t_onewire_o_init;
	
	constant ow_init_total 	: integer := 40 * 700;
	constant ow_init_hiZ		: integer := 40 * 600;
	constant ow_init_sample	: integer := 40 * 630;
	
	constant ow_write_total : integer := 40 * 70;
	constant ow_write_hiZ_0 : integer := 40 * 60;
	constant ow_write_hiZ_1 : integer := 40 * 5;
	
	constant ow_read_total 	: integer := 40 * 70;
	constant ow_read_hiZ		: integer := 40 * 5;
	constant ow_read_sample	: integer := 40 * 15; 
	
	constant ow_maxcount		: integer := ow_init_total;

begin

i <= input;
output <= o;



process (clk40)
	type t_owstatus is (idle, init, wr, rd);
	variable state 			: t_owstatus 									:= idle;
	variable bit_now	 		: integer range 8 downto 0 				:= 0;
	variable counter 			: integer range ow_maxcount downto 0 	:= 0;
	variable bits				: integer range 8 downto 1 				:= 1;
begin
	if rising_edge(clk40) then
		if counter < ow_maxcount then counter := counter + 1;
		end if;
		
		case state is
			----------------------------------------------------------------------
			when idle =>
				if i.ext_wr(ctrl_reg) = '1' then 
					counter := 0;
					bit_now := 0;
					if i.reg(ctrl_reg)(0) = '1' then
						state := init;
					elsif i.reg(ctrl_reg)(1) = '1' then 
						state := wr;
						bits := 1;
					elsif i.reg(ctrl_reg)(2) = '1' then 
						state := rd;
						bits := 1;
					elsif i.reg(ctrl_reg)(3) = '1' then 
						state := wr;
						bits := 8;
					elsif i.reg(ctrl_reg)(4) = '1' then 
						state := rd;
						bits := 8;
					end if;
				end if;
			----------------------------------------------------------------------
			when init =>
				if counter = 1 then wire <= '0';
				elsif counter = ow_init_hiZ then wire <= 'Z';
				elsif counter = ow_init_sample then 
					o.reg_int(read_reg) <= wire & "0000000";
					o.reg_ch_int(read_reg) <= not i.reg_ch_int(read_reg);
					o.reg_int(crc_reg) <= (others => '0');
					o.reg_ch_int(crc_reg) <= not i.reg_ch_int(crc_reg);				
				elsif counter = ow_init_total then state := idle;
				end if;
			----------------------------------------------------------------------
			when wr =>
				if bit_now < bits then
					if counter = 1 then wire <= '0';
					elsif (i.reg(write_reg)(bit_now) = '0' and counter = ow_write_hiZ_0) or (i.reg(write_reg)(bit_now) = '1' and counter = ow_write_hiZ_1) then wire <= 'Z';
					elsif counter = ow_write_total then 
						bit_now := bit_now + 1;
						counter := 0;
					end if;
				else state := idle;
				end if;
			----------------------------------------------------------------------
			when rd =>
				if bit_now < bits then
					if counter = 1 then wire <= '0';
					elsif counter = ow_read_hiZ then wire <= 'Z';
					elsif counter = ow_read_sample then
						o.reg_int(read_reg) <= wire & i.reg(read_reg)(7 downto 1);
						o.reg_ch_int(read_reg) <= not i.reg_ch_int(read_reg);
						o.reg_int(crc_reg) <= (i.reg(crc_reg)(0) xor wire) & 
														(i.reg(crc_reg)(7) )& 
														(i.reg(crc_reg)(6) )& 
														(i.reg(crc_reg)(5) )& 
														(i.reg(crc_reg)(4) xor (i.reg(crc_reg)(0) xor wire) )& 
														(i.reg(crc_reg)(3) xor (i.reg(crc_reg)(0) xor wire) )& 
														(i.reg(crc_reg)(2) )& 
														(i.reg(crc_reg)(1) );
						o.reg_ch_int(crc_reg) <= not i.reg_ch_int(crc_reg);				
					elsif counter = ow_read_total then 
						bit_now := bit_now + 1;
						counter := 0;
					end if;
				else state := idle;
				end if;
		end case;
	end if;
end process;



end Behavioral;

