----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    10:33:14 10/18/2012 
-- Design Name: 
-- Module Name:    rosj3 - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_logIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

library WORK;
use WORK.common.ALL;

entity rosj3 is
	Port (
		input  		: in t_rosj3_i;
		output 		: out t_rosj3_o;
		clk40 		: in std_logic
		);
end rosj3;

architecture Behavioral of rosj3 is

	signal i	: t_rosj3_i;
	signal o : t_rosj3_o := t_rosj3_o_init;

	signal fifo_reset	: std_logic 							:= '1'; -- para vaciar la fifo cuando haya un gsr
	signal fifo_wr		: std_logic 							:= '0';
	signal fifo_rd		: std_logic 							:= '0';
	signal fifo_din	: std_logic_vector(19 downto 0)	:= (others => '0');
	signal fifo_dout	: std_logic_vector(19 downto 0);
	signal fifo_full	: std_logic;
	signal fifo_empty	: std_logic;
	
	signal debugmem_wr	: std_logic_vector(0 downto 0)	:= "0";
	signal debugmem_addr	: unsigned(11 downto 0)				:= (others => '0');
	signal debugmem_din 	: std_logic_vector(31 downto 0)	:= (others => '0');
	signal debugmem_dout	: std_logic_vector(7 downto 0) 	:= (others => '0'); 

	signal fresh : unsigned(fresh_bits-1 downto 0) := (others => '1');

	constant idleword 	: std_logic_vector(19 downto 0) := "00000000001111111111";
	constant ecrword		: std_logic_vector(19 downto 0) 	:= "0" & "00110" & "0000000000000" & "1";
	constant refcount 	: std_logic_vector(4 downto 0) 	:= "00111";
	constant header		: std_logic_vector(3 downto 0) 	:= "0001";
	constant data   		: std_logic_vector(1 downto 0) 	:= "01";
	constant data2  		: std_logic_vector(1 downto 0) 	:= "10";
	constant trailer  	: std_logic_vector(3 downto 0) 	:= "0010";

	COMPONENT tsc_fifo
	  PORT (
		 rst : IN STD_LOGIC;
		 wr_clk : IN STD_LOGIC;
		 rd_clk : IN STD_LOGIC;
		 din : IN STD_LOGIC_VECTOR(19 DOWNTO 0);
		 wr_en : IN STD_LOGIC;
		 rd_en : IN STD_LOGIC;
		 dout : OUT STD_LOGIC_VECTOR(19 DOWNTO 0);
		 full : OUT STD_LOGIC;
		 empty : OUT STD_LOGIC
	  );
	END COMPONENT;

	COMPONENT debugmem
	  PORT (
		 clka : IN STD_LOGIC;
		 wea : IN STD_LOGIC_VECTOR(0 DOWNTO 0);
		 addra : IN STD_LOGIC_VECTOR(9 DOWNTO 0);
		 dina : IN STD_LOGIC_VECTOR(31 DOWNTO 0);
		 clkb : IN STD_LOGIC;
		 addrb : IN STD_LOGIC_VECTOR(11 DOWNTO 0);
		 doutb : OUT STD_LOGIC_VECTOR(7 DOWNTO 0)
	  );
	END COMPONENT;


begin
i <= input;
output <= o;

	o.debug(1)  <= '0';
	o.debug(2)  <= i.word(18);
	o.debug(3)  <= '0';
	o.debug(4)  <= i.word(17);
	o.debug(5)  <= '0';
	o.debug(6)  <= not fifo_empty;
	o.debug(7)  <= '0';
	o.debug(8)  <= '0';
	o.debug(9)  <= '0';
	o.debug(10) <= i.l1a;
	o.debug(11) <= '0';
	o.debug(12) <= '0';
	o.debug(13) <= '0';
	o.debug(14) <= i.ros_ready;
	o.debug(15) <= '0';
	o.debug(16) <= '0';
	o.debug(17) <= '0';
	o.debug(18) <= o.trg_strobe;
--	o.debug(19) <= reset;
--	o.debug(20) <= clk40;


-- clk40 is >= than clk40s, so any operation that only needs to see the input can be done on that clock.
-- here it only writes to the fifo when something that has to go there arrives. 
-- refcount words are written to mark te point where normal operation is restored
process (i.clk40s)
begin
	if rising_edge(i.clk40s) then
		fifo_din <= i.word;  
		-- write fifo if word is not idle... (if in error, will be reset next cycle)
		if i.word(18 downto 1) /= idleword(18 downto 1) and 
			i.word(18 downto 1) /= ecrword(18 downto 1) and 
			i.word(18 downto 14) /= refcount then 
				fifo_wr <= '1';
		else fifo_wr <= '0';
		end if;

	end if;
end process;

tsc_fifo_inst : tsc_fifo
	PORT MAP (
		rd_clk => clk40,
		wr_clk => i.clk40s,
		rst => fifo_reset,
		din => fifo_din,
		wr_en => fifo_wr,
		rd_en => fifo_rd,
		dout => fifo_dout,
		full => fifo_full,
		empty => fifo_empty
		);

process (clk40)
	-- hay muchos contadores de event counter segn dnde se miren. empiezan por ec_
	variable ec_next_l1a 	: unsigned(11 downto 0) 			:= (others => '0'); -- old eventcnt; indicates the ec that will correspond to the next l1a to be received
	variable ec_next_ros		: unsigned(11 downto 0) 			:= (others => '0'); -- old next_ec; indicates the ec that will correspond to the next ros_ready
	variable ec_resync		: unsigned(11 downto 0) 			:= (others => '0'); -- indicates the ec at which the reception will be resynced (the first normal event)
	variable ec_last_hdr_rd : unsigned(11 downto 0) 			:= (others => '1'); -- old last_ec_read_hdr; indicates the last ec whose header has been read from the fifo
	variable last_rd_was_trailer :  std_logic						:= '0'; -- old last_ec_read_complete; indicates last word read from the fifo was a trailer
	variable ec_hdr_rcvd 	: unsigned(11 downto 0)				:= (others => '1'); -- old headerevent; indicates the last header received (fifo input)
	variable ec_half_rcvd 	: unsigned(11 downto 0)				:= (others => '1'); -- old halfevent: indicates the last event received that had ended trg_strobe at origin
	variable ec_tra_rcvd 	: unsigned(11 downto 0)				:= (others => '1'); -- old fullevent: indicates the last event completely received and written to the fifo
	variable ec_offset 		: unsigned(11 downto 0)				:= (others => '0'); -- indicates the difference between the tsc_rear and ofcu_counters
	
	variable latency_counter	: unsigned(7 downto 0)	:= (others => '0');

	variable resync_fifo : std_logic := '0';
	variable resync_ros : std_logic := '0';
	
	variable l1a_pipe : std_logic_vector(3 downto 0) 	:= (others => '0');
	variable sincelastl1a : integer range 255 downto 0 := 0;

	type status_t is (error, resetting, simecr, resync, idle, waiting_ros, waiting_header, waiting_data, passing_data, finishing);
	variable status 			: status_t 							:= idle;
	variable status_old		: status_t 							:= idle;
	variable counter 			: integer range 255*64 downto 0 	:= 0;
	variable changed_status	: std_logic							:= '0';

	variable error_register 		: std_logic_vector(23 downto 0) := (others => '0');
	variable error_register_la 	: std_logic_vector(23 downto 0) := (others => '0');
	variable inerror_live			: std_logic;
	variable last_rosready			: std_logic;
                                             
	constant recordingmax 			: integer := 4;
	variable recording 				: integer range recordingmax downto 0 := 0;
	variable last_empty				: std_logic := '1';
begin
	if rising_edge(clk40) then
		-- Counter that reaches zero a fixed time after init/GSR. 
		if fresh > 0 then fresh <= fresh - 1;
		end if;
	
		-- detectores inmediatos de error
		error_register(err_ixo_lcl) 		:= i.ixo_on;
		error_register(err_bufpll) 		:= not i.bufpll_locked;
		error_register(err_pll) 			:= not i.pll_locked;
		error_register(err_stayatlimit) 	:= i.iodelay_at_limit;
		error_register(err_fifo) 			:= fifo_full;
		error_register(err_lock) 			:= not i.deser_lock;
		if i.deser_lock = '1' and i.word = ecrword and status /= resetting and status_old /= resetting then error_register(err_ecr_rmt) := '1';
		end if;
		
		----------------------------
		-- EVENT COUNTERS CONTROL --
		----------------------------
		-- ec_next_l1a: ECR & L1A
		if i.evcntres = '1' or i.ext_wr(tsc_status_0_la) = '1' or fresh = 1 then 
			fifo_reset <= '1';
			if i.evcntres = '1' then status := resetting;
			else status := simecr;
			end if;
		elsif l1a_pipe(3 downto 2) = "01" then 
			ec_next_l1a := ec_next_l1a + 1;
			sincelastl1a := 0;
		elsif sincelastl1a < 255 then
			sincelastl1a := sincelastl1a + 1;
		end if;
		l1a_pipe := l1a_pipe(l1a_pipe'high-1 downto 0) & i.l1a;
		
		-- ec_next_ros: increased when finished sending an event to the ROS (either normal or error)
		if i.ros_ready = '0' and last_rosready = '1' then ec_next_ros := ec_next_ros + 1;
		end if;
		
		-- ec_hdr_rcvd, ec_half_rcvd, ec_tra_rcvd (received at the input of the fifo)
		if i.word(18 downto 15) = header then ec_hdr_rcvd := unsigned(i.word(12 downto 1));
		elsif i.word(18 downto 17) = data2 then ec_half_rcvd := ec_hdr_rcvd;
		elsif i.word(18 downto 15) = trailer then ec_tra_rcvd := ec_hdr_rcvd;
		end if;
		
		-- ec_last_hdr_rd & last_rd_was_trailer: events that have been read out of the fifo
		if fifo_rd = '1' and fifo_dout(18 downto 15)= header then 
			ec_last_hdr_rd := unsigned(fifo_dout(12 downto 1));
			last_rd_was_trailer := '0';
		elsif fifo_rd = '1' and fifo_dout(18 downto 15) = trailer then
			last_rd_was_trailer := '1';
		end if;
		-------------------------------------
		-- Finished EVENT COUNTERS CONTROL --
		-------------------------------------
		
		-- if going into a new status, reset the counter to 0
		changed_status := '0';
		if status /= status_old then 
			counter := 0;
			changed_status := '1';
		end if;
		status_old := status;

		case status is
			when error => 
				debugmem_din(31 downto 28) <= "0000";
			when resetting =>
				debugmem_din(31 downto 28) <= "0001";
			when simecr =>
				debugmem_din(31 downto 28) <= "0010";
			when resync =>
				debugmem_din(31 downto 28) <= "0011";
			when idle =>
				debugmem_din(31 downto 28) <= "0100";
			when waiting_ros =>
				debugmem_din(31 downto 28) <= "0101";
			when waiting_header =>
				debugmem_din(31 downto 28) <= "0110";
			when waiting_data =>
				debugmem_din(31 downto 28) <= "0111";
			when passing_data =>
				debugmem_din(31 downto 28) <= "1000";
			when finishing =>
				debugmem_din(31 downto 28) <= "1001";
			when others =>
				debugmem_din(31 downto 28) <= "1010";
		end case;
		
		-- state machine
		inerror_live := '0';
		case status is 
			when error => -- empty word passed to the ros ASAP, flush the fifo (but not reset, no need to hurry). 
				fifo_rd <= '1';
				o.trg_data <= (others => '0');
				o.trg_strobe <=  i.ros_ready and not last_rosready;
				inerror_live := '1';
				error_register(err_inerrorlatch) := '1';
				
				if sincelastl1a > to_integer(unsigned(i.reg(tsc_resyncgap))) and i.word(18 downto 14) = refcount then
					ec_offset :=  ec_next_l1a - unsigned(i.word(12 downto 1));
					ec_resync := ec_next_l1a;
					status := resync;
					fifo_rd <= '0';
					resync_fifo := '0';
					resync_ros := '0';
				end if;
				
			when resync =>
				inerror_live := '1';
				if ec_last_hdr_rd+ec_offset+1 = ec_resync and last_rd_was_trailer = '1' then
					fifo_rd <= '0';
					resync_fifo := '1';
				else fifo_rd <= '1';
				end if;
				
				if ec_next_ros /= ec_resync then o.trg_strobe <=  i.ros_ready and not last_rosready;
				else 
					o.trg_strobe <= '0';
					resync_ros := '1';
				end if;
				
				if resync_fifo = '1' and resync_ros = '1' then
					status := idle;
					error_register(err_resynced) := '1';
				elsif counter > 64*to_integer(unsigned(i.reg(tsc_resyncwait))) then
					status := error;
					error_register(err_resync) := '1';
				else counter := counter + 1;
				end if;
				
			when resetting => --wait for serial ecr word, but if timeout, continue like there was no problem, just in case the error is masked
				o.trg_strobe <= '0';
				if i.word = ecrword or counter > to_integer(unsigned(i.reg(tsc_ecrwait))) then 
					status := idle;
					ec_next_l1a := (others => '0');
					ec_next_ros := (others => '0');
					ec_last_hdr_rd := (others => '1');
					ec_offset := (others => '0');
					ec_hdr_rcvd := (others => '1');
					ec_half_rcvd := (others => '1');
					ec_tra_rcvd  := (others => '1');
					fifo_reset <= '0';
					if counter > to_integer(unsigned(i.reg(tsc_ecrwait))) then error_register(err_ecr_lcl) := '1';
					end if;
				else counter := counter + 1;
				end if;
				
			when simecr => -- go to idle after 2 cycles.
				o.trg_strobe <= '0';
				if counter < 2 then
					counter := counter + 1;
				else
					status := idle;
					ec_next_l1a := (others => '0');
					ec_next_ros := (others => '0');
					ec_offset := (others => '0');
					ec_last_hdr_rd := (others => '1');
					ec_hdr_rcvd := (others => '1');
					ec_half_rcvd := (others => '1');
					ec_tra_rcvd  := (others => '1');
					fifo_reset <= '0';
				end if;
				
			when idle => -- start event processing if an un-processed L1A present
				fifo_reset <= '0';
				if ec_next_l1a /= ec_next_ros then status := waiting_ros;
				end if;
				
			when waiting_ros => -- continue when ros is ready. if counter reached, mark it.
				if i.ros_ready = '1' then status := waiting_header;
				elsif counter = 64*to_integer(unsigned(i.reg(tsc_roswait1))) then error_register(err_ros_rdy) := '1';
				else counter := counter + 1;
				end if;
				
			when waiting_header => -- most errors here must lead to error status
				if i.ros_ready = '0' then error_register(err_ros_rdy_fast) := '1';
				elsif fifo_empty = '0' and fifo_dout(18 downto 15)= header then
					if (unsigned(fifo_dout(12 downto 1))+ec_offset) = ec_next_ros then
						fifo_rd <= '1';
						latency_counter := (others => '0');
						status := waiting_data;
					else 
						error_register(err_hdr_ali) := '1';
						o.trg_data <= (others => '0');
						o.trg_strobe <=  '1';
					end if;
				elsif fifo_empty = '0' then 
					error_register(err_hdr_bad) := '1';
					o.trg_data <= (others => '0');
					o.trg_strobe <=  '1';
				elsif counter > to_integer(unsigned(i.reg(tsc_headerwait))) then 
					error_register(err_hdr_wait) := '1';
					o.trg_data <= (others => '0');
					o.trg_strobe <= '1';
				else counter := counter + 1;
				end if;
				
			when waiting_data => 
				if fifo_rd = '1' then fifo_rd <= '0';
				elsif i.ros_ready = '0' then error_register(err_ros_rdy_fast) := '1';
				elsif fifo_empty = '0' and	(i.reg(ow1_ctrl_reg)(5) = '0' or  latency_counter >= unsigned(i.reg(tsc_safetylatency))) then --this ensures that you only start passing data if the speed is 800 Mbps or if the event is half or fully received
					fifo_rd <= '1';
					status := passing_data;
				elsif counter > to_integer(unsigned(i.reg(tsc_datawait))) then 
					error_register(err_dat_wait) := '1';
					o.trg_data <= (others => '0');
					o.trg_strobe <= '1';
				else counter := counter + 1;
				end if;
				
				if (ec_next_ros = (ec_half_rcvd+ec_offset) or ec_next_ros = (ec_tra_rcvd+ec_offset) or latency_counter > 0) and latency_counter < 255 then latency_counter := latency_counter + 1;
				end if;

			when passing_data =>
				if i.ros_ready = '0' then error_register(err_ros_rdy_fast) := '1';
				elsif fifo_empty = '1' then 
					error_register(err_dat_sudden_end) := '1';
					o.trg_data <= (others => '0');
					o.trg_strobe <= '0';
				elsif counter > to_integer(unsigned(i.reg(tsc_maxwords))) then 
					error_register(err_dat_count_max) := '1';
					o.trg_strobe <= '0';
				elsif fifo_dout(18 downto 17) = data or fifo_dout(18 downto 17) = data2 then
					o.trg_data <= fifo_dout(16 downto 1);
					o.trg_strobe <= '1';
					counter := counter + 1;
				elsif fifo_dout(18 downto 15) = trailer then
					o.trg_strobe <= '0';
					fifo_rd <= '0';
					status := finishing;
					if counter /= to_integer(unsigned(fifo_dout(8 downto 1))) then error_register(err_dat_count_match) := '1';
					end if;
				else error_register(err_bad_data_trail) := '1';
				end if;
			
			when others =>
--				fifo_rd <= '0';
--				if fifo_empty = '0' and fifo_dout(18 downto 14) = refcount then fifo_rd <= '1';
				if i.ros_ready = '0' then status := idle;
				elsif counter = to_integer(unsigned(i.reg(tsc_roswait2))) then error_register(err_ros_rdy_held) := '1';
				else counter := counter + 1;
				end if;
		end case;

		-- cancelar salidas e ir a estado de error
		if to_integer(unsigned(error_register and error_mask)) /= 0 then
			status := error;
		end if;
		
		-- actualizar registro de flags de error y latched.
		if i.ext_wr(tsc_status_1_la) = '1' or fresh = 1 then error_register_la := error_register;
		else error_register_la := error_register_la or error_register;
		end if;
		
		o.reg_int(tsc_status_0_la) 	<= error_register_la(7 downto 0);
		o.reg_int(tsc_status_1_la) 	<= error_register_la(15 downto 8);
		o.reg_int(tsc_status_2_la) 	<= inerror_live & error_register_la(22 downto 16);
		o.reg_ch_int(tsc_status_0_la) 	<= not i.reg_ch_int(tsc_status_0_la);
		o.reg_ch_int(tsc_status_1_la) 	<= not i.reg_ch_int(tsc_status_1_la);
		o.reg_ch_int(tsc_status_2_la) 	<= not i.reg_ch_int(tsc_status_2_la);
		
		-- DEBUG MEMORY
		debugmem_din(27 downto 24) <= std_logic_vector(ec_next_ros(1 downto 0)) & i.l1a & i.evcntres;
		debugmem_din(23 downto 20) <= fifo_rd & fifo_empty & last_empty & fifo_reset;
		debugmem_din(19 downto 0) 	<= fifo_dout(19 downto 0);
		
					
		if (recording < recordingmax) and (changed_status = '1' or (fifo_rd = '1' and fifo_empty = '0') or i.l1a='1' or i.evcntres = '1' or recording > 0 or inerror_live = '1') then
			debugmem_wr <= "1";
			debugmem_addr <= (debugmem_addr(debugmem_addr'high downto 2) + 1) & "11";
		else
			debugmem_wr <= "0";
			if i.ext_wr(test0_reg) = '1' then debugmem_addr <= debugmem_addr - 1;
			end if;
		end if;
		last_empty := fifo_empty;

		if (to_integer(unsigned(error_register and error_mask)) /= 0 or inerror_live = '1')  and recording < recordingmax then
			recording := recording + 1;
		end if;
		if i.ext_wr(test1_reg) = '1' or i.ext_wr(tsc_status_0_la) = '1' or fresh = 1 then 
			recording := 0;
			debugmem_wr <= "1";
			debugmem_addr <= (debugmem_addr(debugmem_addr'high downto 2) + 1) & "11";
			debugmem_din <= (others => '1');

		end if;
		
		if fresh > 0 then 
			debugmem_din <= (others => '0');
			debugmem_wr <= "1";
			debugmem_addr <= (debugmem_addr(debugmem_addr'high downto 2) + 1) & "11";
		end if;
		
		o.reg_ch_int(test0_reg) <= not i.reg_ch_int(test0_reg);
		
		last_rosready := i.ros_ready;
		error_register := (others => '0');
	end if;
end process;


debugmem_name : debugmem
  PORT MAP (
    clka => clk40,
    wea => debugmem_wr,
    addra => std_logic_vector(debugmem_addr(debugmem_addr'high downto 2)),
    dina => debugmem_din,
    clkb => clk40,
    addrb => std_logic_vector(debugmem_addr),
    doutb => o.reg_int(test0_reg)
  );


end Behavioral;