----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    08:50:02 03/20/2013 
-- Design Name: 
-- Module Name:    serialclk - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

library WORK;
use WORK.common.ALL;



entity serialclk is
    Port ( input 		: in  t_serialclk_i;
           output 	: out  t_serialclk_o;
			  reset		: in std_logic
			  );
end serialclk;

architecture Behavioral of serialclk is

	signal i	: t_serialclk_i;
	signal o : t_serialclk_o := t_serialclk_o_init;

	-- PLL
	signal CLKFBIN 		: std_logic;
	signal CLKFBOUT 	 	: std_logic;
--	signal CLKFBOUTi 	 	: std_logic;
	signal clk800p 		: std_logic;
	signal clk200p 		: std_logic;
	signal clk80p			: std_logic;
	signal clk40p			: std_logic;
	
	signal clk80			: std_logic;
	signal clk40in			: std_logic;
	signal clkin			: std_logic;
	
	signal selector		: std_logic := '0';


begin
	i <= input;
	output <= o;
	
	selector <= i.reg(ow1_ctrl_reg)(5);
--	selector <= '0';

	clkin_buf : IBUFG
	generic map(IOSTANDARD => "LVCMOS33")
	port map
		(O => clk40in, 
		I => i.clk40pad); 

	clkin <= clk40in;
	

   PLL_BASE_inst : PLL_BASE
   generic map (
      BANDWIDTH => "OPTIMIZED",             -- "HIGH", "LOW" or "OPTIMIZED" 
      CLKFBOUT_MULT => 20,                   -- Multiply value for all CLKOUT clock outputs (1-64)
      CLKFBOUT_PHASE => 0.0,                -- Phase offset in degrees of the clock feedback output
                                            -- (0.0-360.0).
      CLKIN_PERIOD => 25.000,                  -- Input clock period in ns to ps resolution (i.e. 33.333 is 30
                                            -- MHz).
      -- CLKOUT0_DIVIDE - CLKOUT5_DIVIDE: Divide amount for CLKOUT# clock output (1-128)
      CLKOUT0_DIVIDE => 1,
      CLKOUT1_DIVIDE => 4,
      CLKOUT2_DIVIDE => 10,
      CLKOUT3_DIVIDE => 20,
      CLKOUT4_DIVIDE => 1,
      CLKOUT5_DIVIDE => 1,
      -- CLKOUT0_DUTY_CYCLE - CLKOUT5_DUTY_CYCLE: Duty cycle for CLKOUT# clock output (0.01-0.99).
      CLKOUT0_DUTY_CYCLE => 0.5,
      CLKOUT1_DUTY_CYCLE => 0.5,
      CLKOUT2_DUTY_CYCLE => 0.5,
      CLKOUT3_DUTY_CYCLE => 0.5,
      CLKOUT4_DUTY_CYCLE => 0.5,
      CLKOUT5_DUTY_CYCLE => 0.5,
      -- CLKOUT0_PHASE - CLKOUT5_PHASE: Output phase relationship for CLKOUT# clock output (-360.0-360.0).
      CLKOUT0_PHASE => 0.0,
      CLKOUT1_PHASE => 0.0,
      CLKOUT2_PHASE => 0.0,
      CLKOUT3_PHASE => 0.0,
      CLKOUT4_PHASE => 0.0,
      CLKOUT5_PHASE => 0.0,
      CLK_FEEDBACK => "CLKFBOUT",           -- Clock source to drive CLKFBIN ("CLKFBOUT" or "CLKOUT0")
      COMPENSATION => "SYSTEM_SYNCHRONOUS", -- "SYSTEM_SYNCHRONOUS", "SOURCE_SYNCHRONOUS", "EXTERNAL" 
      DIVCLK_DIVIDE => 1,                   -- Division value for all output clocks (1-52)
      REF_JITTER => 0.1,                    -- Reference Clock Jitter in UI (0.000-0.999).
      RESET_ON_LOSS_OF_LOCK => FALSE        -- Must be set to FALSE
   )
   port map (
      CLKFBOUT => CLKFBOUT, -- 1-bit output: PLL_BASE feedback output
      CLKOUT0 => clk800p,
      CLKOUT1 => clk200p,
      CLKOUT2 => clk80p,
      CLKOUT3 => clk40p,
      CLKOUT4 => open,
      CLKOUT5 => open,
      LOCKED => o.pll_locked,     -- 1-bit output: PLL_BASE lock status output
      CLKFBIN => CLKFBIN,   -- 1-bit input: Feedback clock input
      CLKIN => clkin,       -- 1-bit input: Clock input
      RST => reset            -- 1-bit input: Reset input
   );

   BUFG_clk200 : BUFG
   port map (O => o.clk200, I => clk200p );


   BUFGMUX_FB_inst : BUFGMUX
   generic map (CLK_SEL_TYPE => "SYNC")  -- Glitchles ("SYNC") or fast ("ASYNC") clock switch-over
   port map (O => CLKFBIN, I0 => CLKFBOUT, I1 => clk80p, S => selector);
--	CLKFBIN <= o.clk40;

--   BUFG_clk40 : BUFG
--   port map (O => o.clk40, I => clk40p );
--   BUFGMUX_40o_inst : BUFGMUX
--   generic map (CLK_SEL_TYPE => "SYNC")  -- Glitchles ("SYNC") or fast ("ASYNC") clock switch-over
--   port map (O => o.clk40, I0 => clk40p, I1 => clk80p, S => selector);
	o.clk40 <= clk40in;


--   BUFGMUX_40s_inst : BUFGMUX
--   generic map (CLK_SEL_TYPE => "SYNC")  -- Glitchles ("SYNC") or fast ("ASYNC") clock switch-over
--   port map (O => o.clk40s, I0 => clk40p, I1 => clk40p, S => selector);
   BUFG_clk40s : BUFG
   port map (O => o.clk40s, I => clk40p );


	bufpll_gen: for b in 3 downto 0 generate

		BUFPLL_inst : BUFPLL
		generic map (
			DIVIDE => 4,         -- DIVCLK divider (1-8)
			ENABLE_SYNC => TRUE  -- Enable synchrnonization between PLL and GCLK (TRUE/FALSE)
		)
		port map (
			IOCLK => 			o.clk800(b),         -- 1-bit output: Output I/O clock
			LOCK =>				o.bufpll_locked(b),	-- 1-bit output: Synchronized LOCK output
			SERDESSTROBE => 	o.serdesstrobe(b), 	-- 1-bit output: Output SERDES strobe (connect to ISERDES2/OSERDES2)
			GCLK => 				o.clk200,              -- 1-bit input: BUFG clock input
			LOCKED => 			o.pll_locked,          -- 1-bit input: LOCKED input from PLL
			PLLIN => 			clk800p              -- 1-bit input: Clock input from PLL
		);

	end generate;


end Behavioral;

