----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    11:10:08 03/07/2013 
-- Design Name: 
-- Module Name:    statusleds - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

library WORK;
use WORK.common.ALL;

entity statusleds is
    Port ( 
		input : in  t_statusleds_i;
		output : out  t_statusleds_o;
		clk40 : in  STD_LOGIC
		);
end statusleds;

architecture Behavioral of statusleds is
	signal i	: t_statusleds_i;
	signal o : t_statusleds_o := t_statusleds_o_init;

begin
	i <= input;
	output <= o;

	process (clk40)
		-- 2**22 / 40 MHz = 100 ms;
		constant timer_bits				: integer := 22 + 6;
		
		constant postreset_duration	: integer := 22 + 5;
		constant postreset_speed		: integer := 22 + 1; -- less is more
		
		constant slowck_bits				: integer := 22 + 5;
		
		
		
		variable postreset				: std_logic := '1';
		variable timer						: unsigned(timer_bits-1 downto 0) := (others => '0');
		variable sd							: std_logic_vector(3 downto 0);
		variable lock_latch				: std_logic_vector(1 downto 0):= (others => '1');
		variable bseln_latch				: std_logic_vector(1 downto 0):= (others => '1');
		variable l1a_latch				: std_logic_vector(1 downto 0):= (others => '0');
		variable ecr_latch				: std_logic_vector(1 downto 0):= (others => '0');

	begin
		if rising_edge(clk40) then
			timer := timer + 1;
			if postreset = '1' then
				o.leds <= (others => '0');
				o.leds(to_integer(timer(postreset_speed+1 downto postreset_speed))) <= '1';
				if timer(postreset_duration) = '1' then 
					postreset := '0';
					timer := (others => '0');
				end if;
			else
				if timer(slowck_bits) = '1' then timer := (others => '0');
				end if;
				----- LEDS(0): SIGNAL DETECTS
				-- si todos sd ok, continuo on, else mitad del tiempo apagado, 1/8 para cada led, del cual 1/16 significa sd = 1, 1/64 sd = 0
				if timer = 0 then sd := i.reg(sd_status_reg)(3 downto 0);
				elsif sd = "1111" then o.leds(0) <= '1'; -- si todos SD = 1, led encendido
				elsif timer(slowck_bits-1) = '1' then o.leds(0) <= '0'; -- si no, la segunda mitad del tiempo esta apagado
				elsif timer(slowck_bits-4) = '1' then o.leds(0) <= '0'; -- la segunda mitad de cada slot de 1/8 de tiempo est apagado
				elsif timer(slowck_bits-5 downto slowck_bits-6) /= "00" and sd(to_integer(timer(slowck_bits-2 downto slowck_bits-3))) = '0' then o.leds(0) <= '0'; --las ltimas 3/4 partes del tiempo de cada SD sern 0 si el SD est a 0
				else o.leds(0) <= '1'; -- el resto del tiempo, led encendido
				end if;
				
				----- LEDS(1): SLOW CONTROL 
				if timer(slowck_bits-6 downto 0) = 0 then bseln_latch := bseln_latch(0) & i.sc_bsel_n;
				else bseln_latch := bseln_latch(1) & (bseln_latch(0) and i.sc_bsel_n);
				end if;
				o.leds(1) <= not bseln_latch(1) and timer(slowck_bits-6);

				----- LEDS(2): TSC LINK STATUS 
				if timer(slowck_bits-6 downto 0) = 0 then lock_latch := lock_latch(0) & i.deser_lock;
				else lock_latch := lock_latch(1) & 
					-- si est brillando, para la brillar la prxima tiene que mantenerse en todo bien; si est apagado para brillar la proxima basta con que est enganche en algn momento
					((lock_latch(1) and (lock_latch(0) and i.deser_lock)) or (not lock_latch(1) and (lock_latch(0) or i.deser_lock)));
				end if;
				o.leds(2)<= lock_latch(1);
				
				----- LEDS(3): L1A/ECR 
				if timer(slowck_bits-6 downto 0) = 0 then  l1a_latch := l1a_latch(0) & i.l1a;
				else  l1a_latch := l1a_latch(1) & (l1a_latch(0) or i.l1a);
				end if;
				if timer(slowck_bits-3 downto 0) = 0 then  ecr_latch := ecr_latch(0) & i.evcntres;
				else ecr_latch := ecr_latch(1) & (ecr_latch(0) or i.evcntres);
				end if;
				o.leds(3) <= (l1a_latch(1) and timer(slowck_bits-6)) or (ecr_latch(1) and timer(slowck_bits-3));
				
				
				
				----- EXCEPTION FOR TSC_REAR
				if board_type = tscrear then 
					o.leds(0) <= '0';
					o.leds(2) <= '0';
				end if;
				
			end if;
		end if;
	
	end process;


end Behavioral;

